use std::cell::Ref;

use mlua::prelude::*;

pub mod association;
pub mod bit_mask;
mod cast;
pub mod libffi_helper;

pub use self::cast::num_cast;

pub trait FfiSize {
    fn get_size(&self) -> usize;
}

pub trait FfiSignedness {
    fn get_signedness(&self) -> bool {
        false
    }
}

// Provide type conversion between luavalue and ffidata types
pub trait FfiConvert {
    // Write LuaValue into FfiData
    unsafe fn value_into_data<'lua>(
        &self,
        lua: &'lua Lua,
        offset: isize,
        data_handle: &Ref<dyn FfiData>,
        value: LuaValue<'lua>,
    ) -> LuaResult<()>;

    // Read LuaValue from FfiData
    unsafe fn value_from_data<'lua>(
        &self,
        lua: &'lua Lua,
        offset: isize,
        data_handle: &Ref<dyn FfiData>,
    ) -> LuaResult<LuaValue<'lua>>;
}

pub trait FfiData {
    fn check_boundary(&self, offset: isize, size: usize) -> bool;
    unsafe fn get_pointer(&self) -> *mut ();
    fn is_writable(&self) -> bool;
    fn is_readable(&self) -> bool;
}

pub struct FfiArg {
    pub size: usize,
    pub callback_ref_flag: u8,
}

impl Clone for FfiArg {
    fn clone(&self) -> Self {
        Self {
            size: self.size,
            callback_ref_flag: self.callback_ref_flag,
        }
    }
}

pub struct FfiResult {
    pub size: usize,
}

impl Clone for FfiResult {
    fn clone(&self) -> Self {
        Self { size: self.size }
    }
}
